from os import path, remove, chdir, makedirs, system
from jinja2.environment import Environment
from jinja2 import FileSystemLoader
from hitchtest.test import Test
import subprocess
import inspect
import jinja2
import yaml
import time
import sys
import imp
import sys
import os


class Module(object):
    def __init__(self, filename, settings):
        self.filename = path.realpath(filename)
        self.engine_folder = settings['engine_folder']

        self.name = path.split(self.filename)[1].replace(".test", "")
        self.title = self.name.replace("_", " ").title()
        self.dirname = path.dirname(self.filename)


        if settings is None:
            self.settings = {}
        else:
            self.settings = settings

        env = Environment()
        env.loader = FileSystemLoader(self.engine_folder)
        tmpl = env.get_template(filename)
        self.test_yaml_text = tmpl.render(**self.settings)

        self.tests = []
        module_yaml_as_dict = yaml.load(self.test_yaml_text)

        if len(module_yaml_as_dict) == 1:
            self.multiple_tests = False
            self.tests = [Test(module_yaml_as_dict[0], self.settings, self.filename, self.title)]
        else:
            self.multiple_tests = True
            for i, test_yaml in enumerate(module_yaml_as_dict, 1):
                self.tests.append(
                    Test(test_yaml, self.settings, self.filename, "{} {}".format(self.title, i))
                )



from subprocess import call, check_output, PIPE
import re

def check_output_lines(command):
    return check_output(command).decode('utf8').split('\n')

def return_code_zero(command):
    """Returns True if command called has return code zero."""
    return call(command, stdout=PIPE, stderr=PIPE)

def versiontuple(v):
    return tuple(map(int, (v.split("."))))

def version_comparison(version_check, version):
    match = re.compile("((?:\>|\=|\<)\=?)?(.*)").match(version_check)
    comparison = match[0]
    version_to_check = match[1]

    if match == "":
        return versiontuple(version_to_check) == versiontuple(version)
    elif match == "=":
        return versiontuple(version_to_check) == versiontuple(version)
    elif match == "==":
        return versiontuple(version_to_check) == versiontuple(version)
    elif match == "<=":
        return versiontuple(version_to_check) <= versiontuple(version)
    elif match == "<":
        return versiontuple(version_to_check) < versiontuple(version)
    elif match == ">":
        return versiontuple(version_to_check) > versiontuple(version)
    elif match == ">=":
        return versiontuple(version_to_check) >= versiontuple(version)
    else:
        raise RuntimeError("Couldn't understand {}".format(version_check))




from hitchtest.step import Step

class Scenario(object):
    def __init__(self, yaml_scenario):
        self.steps = []

        for i, yaml_step in enumerate(yaml_scenario, 1):
            self.steps.append(Step(yaml_step, i))

    def to_dict(self):
        return [step.to_dict() for step in self.steps]



from jinja2.environment import Environment
from jinja2 import FileSystemLoader
from hitchtest import utils
from os import path
import colorama
import json
import sys


TEMPLATE_DIR = path.join(path.dirname(path.realpath(__file__)), "templates")

class TestPosition(object):
    SETUP = 1
    TEARDOWN = 2
    ON_FAILURE = 3
    ON_SUCCESS = 4
    STEP = 5

class HitchStacktrace(object):
    """Representation of a python stacktrace."""

    def __init__(self, test, where, step=None):
        """Create this object with sys.exc_info()[2]"""
        self.tracebacks = []
        self.test = test
        self.where = where
        self.step = step
        tb_id = 0
        tb = sys.exc_info()[2]
        self.exception = sys.exc_info()[1]

        # Create list of tracebacks
        while tb is not None:
            filename = tb.tb_frame.f_code.co_filename
            if "hitchtest/" not in filename:
                self.tracebacks.append(HitchTraceback(tb_id, tb))
                tb_id = tb_id + 1
            tb = tb.tb_next

    def to_template(self, template="stacktrace_default.jinja2"):
        env = Environment()
        env.loader = FileSystemLoader(TEMPLATE_DIR)
        tmpl = env.get_template(path.basename(template))
        return tmpl.render(
            stacktrace=self.to_dict(),
            TestPosition=TestPosition,
            json=self.to_json(),
            Fore=colorama.Fore,
            Back=colorama.Back,
            Style=colorama.Style,
        )

    def to_json(self):
        return json.dumps(self.to_dict())

    def to_dict(self):
        return {
            'test': self.test.to_dict(),
            'step': self.step.to_dict() if self.step else None,
            'where': self.where,
            'tracebacks': [traceback.to_dict() for traceback in self.tracebacks],
            'exception': str(self.exception),
            'exception_type': "{}.{}".format(
                type(self.exception).__module__, type(self.exception).__name__
            ),
        }

    def __len__(self):
        return len(self.tracebacks)

    def __getitem__(self, index):
        return self.tracebacks[index]



class HitchTraceback(object):
    """Representation of a python traceback caused by a failed test case."""

    def __init__(self, tb_id, traceback):
        self.tb_id = tb_id
        self.traceback = traceback

    def to_dict(self):
        return {
            'id': self.tb_id,
            'filename': self.filename(),
            'lineno': self.lineno(),
            'function': self.func(),
            'line': self.loc(),
            'loc_before': self.loc_before(),
            'loc_after': self.loc_after(),
        }

    def filename(self):
        return self.traceback.tb_frame.f_code.co_filename

    def lineno(self):
        return self.traceback.tb_lineno

    def func(self):
        return self.traceback.tb_frame.f_code.co_name

    def localvars(self):
        return self.traceback.tb_frame.f_locals

    def globalvars(self):
        return self.traceback.tb_frame.f_globals

    def frame(self):
        return self.traceback.tb_frame

    def ipython(self):
        utils.ipython(
            message="Entering {} at line {}".format(self.filename(), self.lineno()),
            frame=self.frame(),
        )

    def loc_before(self):
        with open(self.filename(), 'r') as source_handle:
            contents = source_handle.read().split('\n')
        return [x for x in enumerate(contents[self.lineno() - 3:self.lineno() - 1], self.lineno() - 2)]

    def loc(self):
        with open(self.filename(), 'r') as source_handle:
            contents = source_handle.read().split('\n')
        return contents[self.lineno() - 1]

    def loc_after(self):
        with open(self.filename(), 'r') as source_handle:
            contents = source_handle.read().split('\n')
        return [x for x in enumerate(contents[self.lineno():self.lineno() + 2], self.lineno() + 1)]

    def __repr__(self):
        return "[{}] File {}, line {} in {}: {}".format(
            self.tb_id,
            self.filename(),
            self.lineno(),
            self.func(),
            self.loc()
        )



from hitchtest.environment import checks

class HitchEnvironmentCheckUnavailableException(Exception):
    pass

def verify(rules):
    for rule in rules:
        key = list(rule.keys())[0]
        value = list(rule.values())[0]

        if hasattr(checks, key):
            getattr(checks, key)(value)
        else:
            raise HitchEnvironmentCheckUnavailableException("'{}' not available!".format(key))



from hitchtest.hitch_stacktrace import HitchStacktrace, TestPosition
from hitchtest.utils import log, warn
from hitchtest.scenario import Scenario
from hitchtest.result import Result
from hitchtest.environment import verify
from os import path
import inspect
import time
import imp


class Test(object):
    def __init__(self, yaml_test, settings, filename, title):
        self.settings = settings
        self.filename = filename
        self.name = yaml_test['name'] if 'name' in yaml_test else title
        self.engine = yaml_test['engine']
        self.description = yaml_test['description'] if 'description' in yaml_test else None
        self.preconditions = yaml_test['preconditions'] if 'preconditions' in yaml_test else None
        self.tags = yaml_test['tags'] if 'tags' in yaml_test else None
        self.features = yaml_test['features'] if 'features' in yaml_test else None
        self.scenario = Scenario(yaml_test['scenario'])

        if "test" in self.name.lower():
            warn("WARNING: The word 'test' should not appear in your test name - it is redundant.\n")
        if self.name[0].isdigit():
            raise RuntimeError("ERROR : Test names cannot start with a digit.")

    def to_dict(self):
        return {
            "filename": self.filename,
            "name": self.name,
            "engine": self.engine,
            "preconditions": self.preconditions,
            "description": self.description,
            "tags": self.tags,
            "features": self.features,
            "scenario": self.scenario.to_dict(),
        }

    def run(self):
        start_time = time.time()

        module_source = path.abspath(path.join(self.settings['engine_folder'], self.engine.split(":")[0]))
        engine_class_name = self.engine.split(":")[1]
        engine_module = imp.load_source("engine", module_source)
        engine_class = [x for x in inspect.getmembers(engine_module) if x[0] == engine_class_name][0][1]
        engine = engine_class(self.settings, self.preconditions)
        engine.name = self.name
        engine.aborted = False
        engine._test = self
        failure = False

        log("RUNNING TEST {}\n".format(self.name))
        stacktrace = None

        try:
            verify(self.settings.get("environment", []))
            engine.set_up()
        except Exception as e:
            stacktrace = HitchStacktrace(self, TestPosition.SETUP)
            failure = True

        if not failure:
            for step in self.scenario.steps:
                try:
                    step.run(engine)
                except Exception as e:
                    stacktrace = HitchStacktrace(self, TestPosition.STEP, step=step)
                    failure = True
                    break

        if not engine.aborted:
            if failure:
                try:
                    engine.on_failure(stacktrace)
                except Exception as e:
                    stacktrace = HitchStacktrace(self, TestPosition.ON_FAILURE)
            else:
                try:
                    engine.on_success()
                except Exception as e:
                    stacktrace = HitchStacktrace(self, TestPosition.ON_SUCCESS)

        try:
            engine.tear_down()
        except Exception as e:
            stacktrace = HitchStacktrace(self, TestPosition.TEARDOWN)

        duration = time.time() - start_time
        dict_stacktrace = stacktrace.to_dict() if stacktrace else None
        result = Result(self, failure, duration, stacktrace=dict_stacktrace, aborted=engine.aborted)
        return result



"""Command line interface to hitchtest."""
from click import command, group, argument, option
from hitchtest.utils import log, warn, signals_trigger_exit
from sys import exit, executable
from os import path, walk, chdir
from hitchtest import module
from hitchtest import suite
import yaml as pyyaml
import fnmatch
import signal
import json


# TODO : Remove --results and put the setting in to settings instead.

@command()
@argument('filenames', nargs=-1)
@option('-y', '--yaml', is_flag=True, help='Output the YAML tests (for debugging).')
@option('-q', '--quiet', is_flag=True, help='Quiet mode. Do not print test output to screen.')
@option('-r', '--results', default=None, help='Specify a template to display test results with.')
@option('-s', '--settings', default=None, help="Load settings from specified file.")
@option('-e', '--extra', default=None, help="""Load extra settings on command line as JSON (e.g. --extra '{"postgres_version": "3.5.5"}'""")
def cli(filenames, yaml, quiet, results, settings, extra):
    """Run test files or entire directories containing .test files."""
    # .hitch/virtualenv/bin/python <- 4 directories up from where the python exec resides
    engine_folder = path.abspath(path.join(executable, "..", "..", "..", ".."))
    settings_dict = {'engine_folder': engine_folder, 'quiet': quiet, }
    filenames = [path.relpath(filename, engine_folder) for filename in filenames]
    chdir(engine_folder)
    default_settings_filename = path.join(engine_folder, "settings.yml")
    settings_filename = default_settings_filename if settings is None else path.join(engine_folder, settings)

    if settings is not None and not path.exists(settings_filename):
        warn("Settings file '{}' could not be found!\n".format(settings_filename))
        sys.exit(1)

    # Load settings from file, if it exists
    if path.exists(settings_filename):
        with open(settings_filename) as settingsfile_handle:
            settings_dict.update(pyyaml.load(settingsfile_handle.read()))

    # Load extra settings from command line JSON
    if extra is not None:
        settings_dict.update(json.loads(extra).items())


    # Get list of files from specified files/directories
    matches = []
    for filename in filenames:
        if path.isdir(filename):
            for root, dirnames, filenames_in_dir in walk(filename):
                for filename_in_dir in fnmatch.filter(filenames_in_dir, '*.test'):
                    matches.append(path.join(root, filename_in_dir))
        else:
            matches.append(filename)

    # Get list of modules from matching directly specified files from command line
    # and indirectly (in the directories of) directories specified from cmd line
    test_modules = []
    for filename in matches:
        if filename.endswith(".test"):
            test_modules.append(module.Module(filename, settings_dict))
        else:
            warn("I didn't understand {}\n".format(filename))
            exit(1)

    test_suite = suite.Suite(test_modules, settings_dict)

    if yaml:
        test_suite.printyaml()
    else:
        returned_results = test_suite.run(quiet=quiet)
        log(returned_results.to_template(template=results))
        exit(1 if len(returned_results.failures()) > 0 else 0)

def run():
    """Run hitch tests"""
    signals_trigger_exit()
    cli()

if __name__ == '__main__':
    run()



import inspect
import signal
from hitchtest import utils
from hitchtest.signal_manager import SignalManager

class HitchAbortedException(Exception):
    pass

class ExecutionEngine(object):
    def __init__(self, settings, preconditions):
        self.settings = settings
        self.preconditions = preconditions
        self.signal_manager = SignalManager(self.abort)

    def set_up(self):
        pass

    def on_success(self):
        pass

    def on_failure(self):
        pass

    def tear_down(self):
        pass

    def ipython(self, message=None):
        self.signal_manager.attach_handler(signal.SIGINT, signal.default_int_handler)
        utils.ipython_embed(message)
        self.signal_manager.attach_handler(signal.SIGINT, self.abort)

    def abort(self, signal, frame):
        self.aborted = True
        utils.stop_ipython()
        raise HitchAbortedException
